;======================================================================;
; GO_SECTION --- skip to requested section                             ;
;----------------------------------------------------------------------;
; STATUS: gra 14-Aug-03                       CREATED: gra 02-Apr-2000 ;
;----------------------------------------------------------------------;
; INPUT:                                                               ;
;      para       |--- paragraph  requested (>=0, optional)            ;
;                      can also be a '$x.x.x.x' string, if desired.    ;
;           The following three should not be set if para is a string: ;
;      sect       |--- section    requested (>=0, optional)            ;
;      subsect    |--- subsection requested (>=0, optional)            ;
;      claus      |--- clause     requested (>=0, optional)            ;
;                 ^ all four missing : goto next $ section just_forward;
;           if any of the above four is -1, search for next section    ;
;           with arbitrary entry at the -1 fields (wildcard search)    ;
;      unit        --- file unit to read in >0                         ;
;                      if empty, use current unit (or standard input)  ;
;      /duplicate   --- optional; if set, duplicate sections allowed   ;
;      /clear,      --- clean up current file state and return         ;
;      /cleanup                                                        ;
;      /nowarn      --- do not warn, only return error code            ;
;      /debug       --- switch on debug mode, use with unit keyword    ;
;----------------------------------------------------------------------;
;   give_paragraph  --- optional, return paragraph  found when wildcard;
;   give_section    --- optional, return section    found when wildcard;
;   give_subsection --- optional, return subsection found when wildcard;
;   give_clause     --- optional, return clause     found when wildcard;
;----------------------------------------------------------------------;
; RESULT:                                                              ;
;      i_warn      --- error code:                                     ;
;                     0 --- no error                                   ;
;                    -1 --- end of file reached                        ;
;                     1 --- paragraph  not found                       ;
;                     2 --- section    not found                       ;
;                     3 --- subsection not found                       ;
;                     4 --- clause     not found                       ;
;                    10 --- already in section  (if duplicate=0)       ;
;----------------------------------------------------------------------;
; NOTES:                                                               ;
;        Section numbers must have the format (max. 4 fields)          ;
;        $12.34.56.78BLANK (max 5 digits in each field). Missing       ;
;        components are equivalent to 0. If all four are missing,      ;
;        the routine searches for the next $ section. A single $ is    ;
;        treated as beginning marker of an input group. A second       ;
;        single $ is treated as end marker.                            ;
;        On error codes 1 to 4 and -1, the file is rewound to the last ;
;        successfully read position. On error code 10, the file pointer;
;        stays just where it was. Sections should be read and stored   ;
;        in sequence [multiple reads of same section number allowed].  ;
;        On first read, the unit argument must be present unless       ;
;        reading from standard input. Subsequent reads from the same   ;
;        file can be done without the unit specifier.                  ;
;        Unless the Fortran 90 version, units are not tracked, so      ;
;        intermittent unit changes must always be accompanied by the   ;
;        correct unit number. Read order is not tracked when switching.;
;        Go_Section is smart when searching for entries: It knows when ;
;        a paragraph read in sequence is larger than the requested one ;
;        and immediately stops searching.                              ;
;----------------------------------------------------------------------;
; MAINTENANCE HISTORY:                                                 ;
; gra 14-Aug-03: skip paragraph detection for just_forward mode;       ;
;                corrected small noisy bug in header parsing           ;
; gra 06-Aug-03: corrected error in just_forward mode; eliminated      ;
;                -2 return code (now behaves like x.x.x.x not found);  ;
;                corrected wrong section detection                     ;
; gra 05-Aug-03: implemented alternative call with '$x.x.x.x' argument ;
; gra 04-Aug-03: adapted to newer version of F90 routine               ;
; gra 02-Apr-00: created from Fortran 90 routine go_section            ;
;======================================================================;
function GO_SECTION, para, sect, subsect, claus, unit=unit, duplicate=duplicate, $
                     clear=clear, cleanup=cleanup, nowarn=nowarn, debug=debugs,  $
                     give_paragraph=give_paragraph, give_section=give_section,   $
                     give_subsection=give_subsection, give_clause=give_clause
common save_block, i_para, i_sect, i_sub, i_clause, current_unit, found_title, debug

routine   = 'GO_SECTION'
ERROR     = "### ERROR "+routine+":"
mx_fields = 4
;---------------------------------;
;     initialize common variables ;
;---------------------------------;
if (keyword_set(clear) or keyword_set(cleanup)) then begin
    i_para       = -99
    i_sect       = -99
    i_sub        = -99
    i_clause     = -99
    current_unit = -99
    found_title  = 0
    return, 0
endif
if (keyword_set(unit) or n_elements(i_para) eq 0) then begin
    debug = 0
    if (keyword_set(debugs)) then debug = 1
    if (debug gt 0) then print, "initializing common"
    i_para       = -99
    i_sect       = -99
    i_sub        = -99
    i_clause     = -99
    current_unit = -99
    found_title  = 0
endif
paragraph    = 0
section      = 0
subsection   = 0
clause       = 0

give_any     = 0
if (arg_present(give_paragraph)  or arg_present(give_section) or $
    arg_present(give_subsection) or arg_present(give_clause)) then give_any = 1
;-----------------------------------------;
;     check for section string or numbers ;
;-----------------------------------------;
type_para = (size(para, /struct)).type
if (type_para eq 0) then begin
    paragraph    = -99
    section      = -99
    subsection   = -99
    clause       = -99
endif else if (type_para eq 7) then begin
    if (strlen(strtrim(para,2)) eq 0) then goto,SKIP_SECT
    if (strmid(para,0,1) ne '$') then begin
        print, ERROR+'Invalid para input >>'+para+'<<, required "$x.x.x.x" !'
        stop
    endif else begin
        sections   = strsplit(strmid(para,1),'.',/extract)
        if (n_elements(sections) gt mx_fields) then begin
            print,ERROR+'Invalid para input >>'+para+'<<, required max 4 fields !'
            stop
        endif
        if (n_elements(sections) gt 0) then paragraph  = sections[0]
        if (n_elements(sections) gt 1) then section    = sections[1]
        if (n_elements(sections) gt 2) then subsection = sections[2]
        if (n_elements(sections) gt 3) then clause     = sections[3]
    endelse
endif else if (type_para ne 2) then begin
    print,ERROR+'Invalid para input type, required character or integer !'
    stop
endif else begin
    if (keyword_set(subsect))  then subsection = subsect
    if (keyword_set   (sect))  then    section = sect
    if (keyword_set   (para))  then paragraph  = para
    if (keyword_set   (claus)) then    clause  = claus
endelse
SKIP_SECT:
;---------------------------------;
;     check for duplicate keyword ;
;---------------------------------;
uniq    = 1
if (keyword_set(duplicate)) then uniq = 1 - duplicate
;----------------------------;
;     check for unit keyword ;
;----------------------------;
i_warn  = 0
f_unit  = current_unit
if (keyword_set(unit)) then f_unit = unit
;----------------------------------------;
;     initialize file_unit on first call ;
;----------------------------------------;
if (current_unit eq -99) then begin
    if (f_unit eq -99) then f_unit = 0
    current_unit =  f_unit
    i_para       = -99
    i_sect       = -99
    i_sub        = -99
    i_clause     = -99
    found_title  = 0
endif
;---------------------------------------------;
;     check if unit changed and save new unit ;
;---------------------------------------------;
if (f_unit ne current_unit) then begin
    current_unit = f_unit
    i_para       = -99
    i_sect       = -99
    i_sub        = -99
    i_clause     = -99
    found_title  = 0
endif
;---------------------------------;
;     initial setup, error checks ;
;---------------------------------;

just_forward = 1
if (paragraph ne -99 or section ne -99 or subsection ne -99 or clause ne -99) then just_forward = 0
if ((paragraph   lt -1  or  paragraph  gt 99999  or section lt -1  or  section  gt 99999  or  $
     subsection  lt -1  or  subsection gt 99999  or clause  lt -1  or  clause   gt 99999) and not just_forward)  then $
     stop,ERROR+"Numbers must be >= 0 and <= 999 !"
wildcards   = 0
if (paragraph eq -1 or section eq -1 or subsection eq -1 or clause eq -1) then wildcards = 1
;-----------------------------------------------------------------;
;     check if already in paragraph (except title and end marker) ;
;-----------------------------------------------------------------;
found_paragraph  = 0
found_section    = 0
found_subsection = 0
found_clause     = 0
if (just_forward eq 0 and wildcards eq 0 and (paragraph ne 0 or section ne 0 or subsection ne 0 or clause ne 0)) then begin
    if (paragraph lt i_para) then print, routine, ' Paragraphs must be read in sequence, paragraph=', paragraph
    if (paragraph eq i_para) then begin
        if (section lt i_sect) then print, routine, ' Sections must be read in sequence, section=', section
        if (section eq i_sect) then begin
            if (subsection lt i_sub) then print, routine,' Subsections must be read in sequence, subsection=',subsection
            if (subsection eq i_sub) then begin
                if (clause lt i_clause) then print, routine,' Clauses must be read in sequence, clause=',clause
                if (clause eq i_clause and uniq eq 1) then begin
                    if (not keyword_set(nowarn)) then print, routine, ' Already in clause ', clause
                    i_warn = 10
                    return, i_warn
                endif
                found_subsection = 1
            endif
            found_section = 1
        endif
        found_paragraph = 1
    endif
endif
;====================;
;     search section ;
;====================;
i_record = 0L
current  = (fstat(f_unit)).cur_ptr
SEARCH_SECTION:
found_sections = [found_paragraph, found_section,found_subsection,found_clause]
while ((where(found_sections eq 0))[0] ne -1 or found_title eq 0) do begin
;-----------------------;
;     look for $ - line ;
;-----------------------;
   head   = ' '
   SEARCH_DOLLAR:
   while (strmid(head,0,1) ne '$') do begin
       if (EOF(f_unit)) then begin
           i_warn = -1
           goto, EXIT_SEARCH_SECTION
       endif
       readf, f_unit, head
       i_record = i_record + 1
   endwhile
;------------------------------------------------;
;     extract paragraph, section, and subsection ;
;------------------------------------------------;
   i_para   = 0
   i_sect   = 0
   i_sub    = 0
   i_clause = 0

   if (just_forward eq 1 and give_any eq 0) then goto, EXIT_SEARCH_SECTION
   head  = (strsplit(strmid(head,1),' ',/extract,/preserve_null))[0]
   
   if (debug gt 0) then print,'head=',head
   s_arr  = strsplit(head,'.',/extract)
   if (strlen(s_arr[0]) eq 0) then begin
       i_para = 0
   endif else begin
       if (n_elements(s_arr) gt mx_fields) then stop, ERROR+'Bad header format:',head

       if (n_elements(s_arr) ne 0) then begin
           i_para = fix(s_arr[0])
           if (n_elements(s_arr) gt 1) then i_sect   = fix(s_arr[1])
           if (n_elements(s_arr) gt 2) then i_sub    = fix(s_arr[2])
           if (n_elements(s_arr) gt 3) then i_clause = fix(s_arr[3])
       endif else begin
           if (just_forward eq 1) then goto, EXIT_SEARCH_SECTION
       endelse
   endelse
;-----------------------------------------------------------------;
;     title and end marker handling (message if unusual sequence) ;
;-----------------------------------------------------------------;
   if (i_para eq 0 and i_sect eq 0 and i_sub eq 0 and i_clause eq 0) then begin
       if (found_title eq 1) then begin
           i_para   = -99
           i_sect   = -99
           i_sub    = -99
           i_clause = -99
           i_warn   =  -2
           if (paragraph eq 0 and section eq 0 and subsection eq 0 and clause eq 0) then begin
               found_title = 0
               if (debug gt 0) then print, routine+":End Marker reached !"
           endif else begin
               if (not keyword_set(nowarn)) then print, routine+": End Marker reached !"
           endelse
           goto, EXIT_SEARCH_SECTION
       endif
       found_title = 1
       if (paragraph ne 0 or section ne 0 or subsection ne 0 or clause ne 0) then begin
           if (debug gt 0) then print, routine+":Title found !"
       endif
   endif
;------------------;
;     set switches ;
;------------------;
   if (just_forward eq 1) then goto, EXIT_SEARCH_SECTION
   sections       = [paragraph,section,subsection,clause]
   i_fields       = [i_para,i_sect,i_sub,i_clause]
   if (wildcards eq 1) then begin
       first_wild = min(sections)
   endif else begin
       first_wild = mx_fields+1
   endelse
;---------------------------------------;
;     early detection of search failure ;
;---------------------------------------;
   FOUND_SCAN: for i=0,mx_fields-1 do begin
       in_section = 0
       if (sections[i] eq i_fields[i] or sections[i] eq -1) then in_section = 1
       if (in_section eq 1) then begin
           found_sections[i] = 1
           if (i eq mx_fields-1) then goto,EXIT_SEARCH_SECTION
       endif else begin
           if (i lt first_wild-1) then begin
               outside = where( i_fields[0:i] lt sections[0:i] )
               if (outside[0] eq -1L) then goto,EXIT_SEARCH_SECTION
           endif
           goto, EXIT_FOUND_SCAN
       endelse
   endfor
   EXIT_FOUND_SCAN:
           
endwhile
EXIT_SEARCH_SECTION:
found_paragraph  = found_sections[0]
found_section    = found_sections[1]
found_subsection = found_sections[2]
found_clause     = found_sections[3]
;-----------------------------------------------------;
;     nothing found, set i_warn parameter accordingly ;
;-----------------------------------------------------;
if (debug gt 0) then print,'i_para, i_sect, i_sub, i_clause=',i_para, i_sect, i_sub, i_clause
if (debug gt 0) then print,'found = ',found_sections

if (just_forward eq 0) then begin

    if (found_paragraph eq 0) then begin
        if (not keyword_set(nowarn)) then print, routine+": Input file, not found paragraph ", paragraph
        i_para   = -99
        i_sect   = -99
        i_sub    = -99
        i_clause = -99
        i_warn   =   1
    endif else if (found_subsection eq 0) then begin
        if (not keyword_set(nowarn)) then begin
            print, routine+":Input file, paragraph ", paragraph
            print, "Not found section     ", section
        endif
        i_sect   = -99
        i_sub    = -99
        i_clause = -99
        i_warn   =   2
    endif else if (found_subsection eq 0) then begin
        if (not keyword_set(nowarn)) then begin
            print, routine, routine+":Input file, paragraph ", paragraph
            print, routine, "Input file, section   ", section
            print, routine, "Not found subsection  ", subsection
        endif
        i_sub    = -99
        i_clause = -99
        i_warn   =   3
    endif else if (found_clause eq 0) then begin
        if (not keyword_set(nowarn)) then begin
            print, routine, routine+":Input file, paragraph  ", paragraph
            print, routine, "Input file, section    ", section
            print, routine, "Input file, subsection ", subsection
            print, routine, "Not found clause       ", clause
        endif
        i_clause = -99
        i_warn   =   4
    endif
endif

if (arg_present(give_paragraph )) then give_paragraph  = i_para
if (arg_present(give_section   )) then give_section    = i_sect
if (arg_present(give_subsection)) then give_subsection = i_sub
if (arg_present(give_clause    )) then give_clause     = i_clause
if (i_warn ne 0) then point_lun,f_unit,current

return, i_warn
end
